
#include "Main.h"

/***************************************************************************\
*																			*
*								Obsluha pamti								*
*																			*
\***************************************************************************/

/////////////////////////////////////////////////////////////////////////////
// obsluha vcevlknovho reimu

#ifdef _MT
#define HEAPSERIALIZE 0					// pro vcevlknov reim je serializace
#else
#define HEAPSERIALIZE HEAP_NO_SERIALIZE	// vypnut serializace, nen-li vcevlknov reim
#endif

#ifdef _MT
static	CRITICAL_SECTION	MemoryCriticalSection;	// kritick sekce pro sprvce pamti
#endif

/////////////////////////////////////////////////////////////////////////////
// konstanty - velikosti mus bt mocnina 2 !

#ifndef _M_ALPHA
#define	PAGESIZE		0x1000			// velikost alokan strnky pro ostatn procesory (4096)
#else
#define	PAGESIZE		0x2000			// velikost alokan strnky pro procesor Alpha (8192)
#endif

#define PAGEITEMS		(PAGESIZE/sizeof(ITEM))	// poet prvk na strnku (512 pop. 1024)
#define	REGPAGES		(PAGESIZE/sizeof(PAGE)) // poet strnek na region (256 pop. 512)
#define DATAPAGES		(REGPAGES-1)			// poet datovch strnek na region (255 pop. 511)
#define REGSIZE			(PAGESIZE*REGPAGES)		// velikost regionu (1 MB pop. 4 MB)
#define NUMREGS			(0x78000000/REGSIZE-1)	// poet region v pamti (2039 pop. 509)
#define REGSYSTEM		(0x800000/REGSIZE)		// poet systmovch region na zatku (8 pop. 2)
#define MAXREGS			((0x40000000/REGSIZE)*4)// poet region celkem (4096 pop. 1024)

/////////////////////////////////////////////////////////////////////////////
// struktury

struct MODUL_;

// struktura datovho prvku - 8 bajt
typedef struct ITEM_
{
	ITEM_*		NextItem;				// adresa dalho volnho prvku (NULL=konec)
	ITEM_*		PrevItem;				// adresa pedelho volnho prvku (NULL=zatek)
} ITEM;

// struktura dat strnky - velikost PAGESIZE
typedef struct DATA_
{
	ITEM		Items[PAGEITEMS];		// data jedn strnky
} DATA;

// struktura popisovae strnky - velikost 16 bajt
typedef struct PAGE_
{
	PAGE_*		NextPage;				// adresa dal strnky (NULL=konec)
	PAGE_*		PrevPage;				// adresa pedel strnky (NULL=zatek)
	MODUL_*		Modul;					// adresa popisovae modulu
	long		Used;					// ta pouitch poloek
} PAGE;

// struktura popisovae modulu
typedef struct MODUL_
{
	ITEM*		FreeItem;				// adresa volnho prvku
	PAGE*		ResPage;				// rezervn strnka (NULL=nen)
	long		Size;					// velikost datov poloky
} MODUL;

// struktura regionu - velikost REGSIZE
typedef struct REGION_
{
	long		Used;					// ta pouitch strnek
	long		res[3];					// rezerva pro zarovnn
	PAGE		Pages[DATAPAGES];		// popisovae strnek
	DATA		Data[DATAPAGES];		// data
} REGION;


/////////////////////////////////////////////////////////////////////////////
// promnn

static	HANDLE	hHeap = NULL;			// handle haldy

// obsluha region
static	char*	UsedRegion;				// buffer pznak pouit region
static	int		NextRegion = REGSYSTEM;	// index pt pidlovanho regionu
static	REGION*	ResRegion = NULL;		// rezervn region (NULL=nen)

// obsluha nepipojench strnek (nsleduje etzec pomoc NextPage)
static	PAGE*	FreePage = NULL;		// prvn nepipojen strnka (NULL=nen)

// popisovae datovch modul
static	MODUL	Modul[12] = {
					NULL,	NULL,	0x0008,
					NULL,	NULL,	0x0010,
					NULL,	NULL,	0x0020,
					NULL,	NULL,	0x0040,
					NULL,	NULL,	0x0080,
					NULL,	NULL,	0x0100,
					NULL,	NULL,	0x0200,
					NULL,	NULL,	0x0400,
					NULL,	NULL,	0x0800,
					NULL,	NULL,	0x1000,
					NULL,	NULL,	0x2000,
					NULL,	NULL,	0x4000 };


/////////////////////////////////////////////////////////////////////////////
// prototypy loklnch funkc

static ITEM* NewItem(MODUL* modul);				// vytvoen novho datovho prvku
static void DelItem(MODUL* modul, ITEM* item);	// zruen datovho prvku
static BOOL NewPage(MODUL* modul);				// vytvoen nov strnky
static void DelPage(MODUL* modul);				// zruen rezervn strnky
static BOOL NewRegion();						// vytvoen novho regionu
static void DelRegion();						// zruen rezervnho regionu


/////////////////////////////////////////////////////////////////////////////
// inicializace sprvce pamti (TRUE=inicializace OK)

BOOL MemInit()
{
// kontrola konstant
	ASSERT(sizeof(ITEM) == 8);
	ASSERT(sizeof(PAGE) == 16);
	ASSERT(sizeof(REGION) == REGSIZE);
	ASSERT((REGSIZE & 0xfffff) == 0);
	ASSERT(PAGESIZE <= 0x4000);

// funkce me bt volna jen jednou
	ASSERT(hHeap == NULL);
	if (hHeap != NULL) return TRUE;

// vytvoen haldy
	hHeap = ::HeapCreate(HEAPSERIALIZE, PAGESIZE, 0);
	if (hHeap == NULL) return FALSE;

// vytvoen bufferu pznak pouit region
	ASSERT(FALSE == 0);
	UsedRegion = (char*) ::HeapAlloc(hHeap, HEAP_ZERO_MEMORY | HEAPSERIALIZE, MAXREGS);
	if (UsedRegion == NULL) return FALSE;

// inicializace uzamykn pstupu k pamti
#ifdef _MT
	::InitializeCriticalSection(&MemoryCriticalSection);
#endif

// vytvoen prvnho regionu (pro zkouku pamti)
	return NewRegion();
}


/////////////////////////////////////////////////////////////////////////////
// ukonen sprvce pamti

void MemTerm()
{
// kontrola, zda je sprvce pamti ji ukonen
	ASSERT(hHeap != NULL);
	if (hHeap == NULL) return;

// uvolnn haldy
	::HeapDestroy(hHeap);
	hHeap = NULL;

// uvolnn uzamykn pstupu k pamti
#ifdef _MT
	::DeleteCriticalSection(&MemoryCriticalSection);
#endif
}


/////////////////////////////////////////////////////////////////////////////
// chyba pamti

void MemErr()
{
	::MessageBox(NULL, _T("Lituji, program bude ukonen z dvodu nedostatku pamti!"),
		_T("Nedostatek pamti"), MB_OK | MB_ICONSTOP);
	Exit(253);
}


/////////////////////////////////////////////////////////////////////////////
// pidlen bloku pamti (povolena 0) NULL=chyba

void* MemGet(int size)
{
	void* result = NULL;

// kontrola zadan velikosti bloku
	ASSERT((size >= 0) && (size < 0x7ffffff0));
	if (size <= 0) size = 1;

// bude pidlen malho bloku
	if (size <= PAGESIZE)
	{

// uzamknut pstupu k pamti
#ifdef _MT
		::EnterCriticalSection(&MemoryCriticalSection);
#endif

// uren modulu velikosti prvku
		MODUL* modul = Modul;		// zatek popisova modul
		while (size > modul->Size)
		{
			modul++;				// zven ukazatele modul
		}

// vytvoen datovho prvku
		result = (void*)NewItem(modul);
		ASSERT(result != NULL);

// odemknut pstupu k pamti
#ifdef _MT
		::LeaveCriticalSection(&MemoryCriticalSection);
#endif
	}

// pidlen bloku z haldy, pokud nebyl blok pidlen
	if (result == NULL)
	{
		result = ::HeapAlloc(hHeap, HEAPSERIALIZE, (size + 0xff) & ~0xff);
	}

	if (result == NULL) MemErr();	// chyba pamti

	return result;
}


/////////////////////////////////////////////////////////////////////////////
// uvolnn bloku pamti (povoleno NULL)

void MemFree(void* adr)
{
// adresa NULL se ignoruje
	if (adr == NULL) return;

// nele-li ve vlastnm regionu, uvolnn pomoc haldy (jet nen nutn uzamknut!)
	if (UsedRegion[(DWORD)adr/REGSIZE] == FALSE)
	{
		::HeapFree(hHeap, HEAPSERIALIZE, adr);
		return;
	}

// uzamknut pstupu k pamti
#ifdef _MT
	::EnterCriticalSection(&MemoryCriticalSection);
#endif

// pprava ukazatel
	REGION* region = (REGION*)((DWORD)adr & ~(REGSIZE-1)); // adresa regionu
	DATA* data = (DATA*)((DWORD)adr & ~(PAGESIZE-1)); // data
	PAGE* page = (PAGE*)((DWORD)region +
		((DWORD)data & (REGSIZE-1)) / (sizeof(DATA)/sizeof(PAGE))); // strnka

// kontrola, zda je strnka platn
	if (page->Used > 0)
	{

// kontrola, zda adresa le na platn hranici
		MODUL* modul = page->Modul;
		if (((DWORD)adr & (modul->Size-1)) == 0)
		{

// uvolnn poloky
			DelItem(modul, (ITEM*)adr);
		}
	}

// odemknut pstupu k pamti
#ifdef _MT
	::LeaveCriticalSection(&MemoryCriticalSection);
#endif
}


/////////////////////////////////////////////////////////////////////////////
// zmna velikosti bloku pamti (povolena adresa NULL i velikost 0, pi chyb nezmnno)
// Pozn.: zde nen nutn uzamykat pam (znan to zpomaluje) !!!

void* MemSize(void* adr, int size)
{
	int oldsize;
	void* newadr;

// pi adrese NULL alokace pamti
	if (adr == NULL)
	{
		if (size <= 0) return NULL;
		return MemGet(size);
	}

// pi velikosti 0 uvolnn bloku
	if (size <= 0)
	{
		MemFree(adr);
		return NULL;
	}

// kontrola, zda blok le ve vlastnm regionu
	if (UsedRegion[(DWORD)adr/REGSIZE] != FALSE)
	{

// pprava ukazatel
		REGION* region = (REGION*)((DWORD)adr & ~(REGSIZE-1)); // adresa regionu
		DATA* data = (DATA*)((DWORD)adr & ~(PAGESIZE-1)); // data
		PAGE* page = (PAGE*)((DWORD)region +
			((DWORD)data & (REGSIZE-1)) / (sizeof(DATA)/sizeof(PAGE))); // strnka

// kontrola, zda je strnka platn
		if (page->Used > 0)
		{

// kontrola, zda adresa le na platn hranici
			MODUL* modul = page->Modul;
			oldsize = modul->Size;

			if (((DWORD)adr & (oldsize-1)) == 0)
			{

// test, zda je poteba velikost mnit
				if (((size <= 8) && (oldsize <= 8)) ||
					((size <= oldsize) && (size > oldsize/4)))
				{
					return adr;					// adresa nezmnna
				}

// vytvoen novho bloku
				newadr = MemGet(size);

// kopie starch dat bloku
				if (newadr != NULL)
				{
					if (size >= oldsize)
					{
						MemCopy(newadr, adr, oldsize);
					}
					else
					{
						MemCopy(newadr, adr, size);
					}
					MemFree(adr);
				}
				return newadr;
			}

// neplatn adresa - chybov nvrat
			else
			{
				return NULL;
			}
		}

// neplatn strnka - chybov nvrat
		else
		{
			return NULL;
		}
	}

// penesen bloku z haldy do malch blok
	if (size <= PAGESIZE/2)
	{
		oldsize = (int)::HeapSize(hHeap, HEAPSERIALIZE, adr);
		if (oldsize == -1)
		{
			MemErr();
			return NULL;
		}

		newadr = MemGet(size);

		if (newadr != NULL)
		{
			if (size >= oldsize)
			{
				MemCopy(newadr, adr, oldsize);
			}
			else
			{
				MemCopy(newadr, adr, size);
			}
			::HeapFree(hHeap, HEAPSERIALIZE, adr);
		}
		return newadr;
	}

// realokace bloku v hald
	newadr = ::HeapReAlloc(hHeap, HEAPSERIALIZE, adr, (size + 0xff) & ~0xff);
	if (newadr == NULL) MemErr();	// chyba pamti
	return newadr;
}


/////////////////////////////////////////////////////////////////////////////
// vytvoen novho datovho prvku

static ITEM* NewItem(MODUL* modul)
{
// zajitn platnho prvku
	if (modul->FreeItem == NULL)
	{
		if (NewPage(modul) == FALSE)
		{
			return NULL;
		}
	}

// uren adresy popisovae strnky
	ITEM* item = modul->FreeItem;		// adresa prvku
	REGION* region = (REGION*)((DWORD)item & ~(REGSIZE-1)); // adresa regionu
	DATA* data = (DATA*)((DWORD)item & ~(PAGESIZE-1)); // data
	PAGE* page = (PAGE*)((DWORD)region +
		((DWORD)data & (REGSIZE-1)) / (sizeof(DATA)/sizeof(PAGE))); // strnka

// inkrementace pouitch promnnch strnky, zruen rezervace strnky
	page->Used++;						// inkrementace pouitch poloek
	if (page == modul->ResPage)			// je to rezervn strnka?
	{
		modul->ResPage = NULL;			// zruen rezervn strnky
	}

// vyazen prvku z etzce volnch prvk
	ITEM* nextitem = item->NextItem;	// adresa dalho prvku
	modul->FreeItem = nextitem;
	if (nextitem != NULL)
	{
		nextitem->PrevItem = NULL;
	}
	return item;
}


/////////////////////////////////////////////////////////////////////////////
// zruen datovho prvku

static void DelItem(MODUL* modul, ITEM* item)
{
// pprava adres popisova
	REGION* region = (REGION*)((DWORD)item & ~(REGSIZE-1)); // adresa regionu
	DATA* data = (DATA*)((DWORD)item & ~(PAGESIZE-1)); // data
	PAGE* page = (PAGE*)((DWORD)region +
		((DWORD)data & (REGSIZE-1)) / (sizeof(DATA)/sizeof(PAGE))); // strnka

// zaazen do etzce volnch prvk
	item->PrevItem = NULL;
	item->NextItem = modul->FreeItem;
	if (item->NextItem != NULL)
	{
		item->NextItem->PrevItem = item;
	}
	modul->FreeItem = item;

// dekrementace pouitch prvk strnky, pp. uvolnn strnky
	page->Used--;
	if (page->Used == 0)
	{
		if (modul->ResPage != NULL)
		{
			DelPage(modul);
		}
		modul->ResPage = page;
	}
}


/////////////////////////////////////////////////////////////////////////////
// vytvoen rezervn strnky

static BOOL NewPage(MODUL* modul)
{
// zajitn pipojiteln strnky
	if (FreePage == NULL)
	{
		if (NewRegion() == FALSE)
		{
			return FALSE;
		}
	}

// uren adresy dat pipojovan strnky
	PAGE* page = FreePage;				// pipojovan strnka
	REGION* region = (REGION*)((DWORD)page & ~(REGSIZE-1)); // adresa regionu
	DATA* data = (DATA*)((DWORD)region + 
		((DWORD)page & (REGSIZE-1)) * (sizeof(DATA)/sizeof(PAGE))); // data

// pipojen dat strnky
	if (::VirtualAlloc(data, PAGESIZE,
			MEM_COMMIT, PAGE_READWRITE) == NULL)
	{
		return FALSE;
	}

// inkremetace pouitch strnek regionu, zruen rezervace regiony
	region->Used++;						// inkrementace pouitch strnek
	if (region == ResRegion)			// je to rezervn region?
	{
		ResRegion = NULL;				// zruen rezervnho regionu
	}

// vyazen strnky z etzce nepipojench strnek
	PAGE* nextpage = page->NextPage;
	FreePage = nextpage;
	if (nextpage != NULL)
	{
		nextpage->PrevPage = NULL;
	}

// napojen do etzce volnch poloek modulu
	int size = modul->Size;				// velikost datov poloky
	ITEM* previtem;
	ITEM* nextitem = data->Items;		// ukazatel adresy prvku
	ITEM* item = NULL;					// nen pedel prvek
	for (int i = PAGESIZE/size; i > 0; i--)
	{
		previtem = item;
		item = nextitem;
		nextitem = (ITEM*)((char*)nextitem + size);
		item->PrevItem = previtem;
		item->NextItem = nextitem;
	}
	item->NextItem = modul->FreeItem;
	modul->FreeItem = data->Items;

// inicializace poloek popisovae strnky
	page->Modul = modul;				// adresa modulu
	page->Used = 0;						// poet pouitch poloek

// nastaven strnky jako rezervn
	modul->ResPage = page;				// bude te jako rezervn strnka
	return TRUE;
}


/////////////////////////////////////////////////////////////////////////////
// zruen rezervn strnky

static void DelPage(MODUL* modul)
{
// kontrola, zda je rezervn strnka
	PAGE* page = modul->ResPage;
	if (page == NULL) return;
	modul->ResPage = NULL;

// pprava adresy regionu a dat
	REGION* region = (REGION*)((DWORD)page & ~(REGSIZE-1)); // adresa regionu
	DATA* data = (DATA*)((DWORD)region + 
		((DWORD)page & (REGSIZE-1)) * (sizeof(DATA)/sizeof(PAGE))); // data

// odpojen z etzce volnch poloek
	int size = modul->Size;
	ITEM* item = data->Items;
	ITEM* nextitem;
	ITEM* previtem;
	for (int i = PAGESIZE/size; i > 0; i--)
	{
		nextitem = item->NextItem;
		previtem = item->PrevItem;

		if (previtem != NULL)
		{
			previtem->NextItem = nextitem;
		}
		else
		{
			modul->FreeItem = nextitem;
		}

		if (nextitem != NULL)
		{
			nextitem->PrevItem = previtem;
		}
		item = (ITEM*)((char*)item + size);
	}

// odpojen strnky
	VERIFY (::VirtualFree(data, PAGESIZE, MEM_DECOMMIT));

// napojen strnky do etzce nepipojench strnek
	page->PrevPage = NULL;
	page->NextPage = FreePage;
	if (FreePage != NULL)
	{
		FreePage->PrevPage = page;
	}
	FreePage = page;
	
// snen tae pouitch strnek regionu, pp. uvolnn regionu
	region->Used--;
	if (region->Used == 0)
	{
		if (ResRegion != NULL)
		{
			DelRegion();		// uvolnn rezervnho regionu
		}
		ResRegion = region;		// je to nyn rezervn region
	}
}


/////////////////////////////////////////////////////////////////////////////
// vytvoen rezervnho regionu

static BOOL NewRegion()
{
// lokln promnn
	REGION* region;						// ukazatel adresy regionu
	PAGE*	page;						// ukazatel adresy strnky
	PAGE*	nextpage;					// pomocn ukazatel - pt strnka
	PAGE*	prevpage;					// pomocn ukazatel - pedchoz strnka
	char*	used;						// ukazatel pouitch region
	int i;								// pomocn ta

// nalezen nepouitho regionu
	for (i = NUMREGS;  i > 0; i--)
	{

// adresa zkouenho regionu
		region = (REGION*)(void*)(NextRegion*REGSIZE); // adresa regionu
		used = &(UsedRegion[NextRegion]);	// pznak pouit regionu

// posun indexu ptho regionu
		NextRegion++;				// zven indexu regionu
		if (NextRegion >= REGSYSTEM+NUMREGS)
		{
			NextRegion = REGSYSTEM;
		}
		
// test, zda je region pouit		
		if (*used == FALSE)
		{

// rezervovn novho regionu
			region = (REGION*)::VirtualAlloc(region,
					REGSIZE, MEM_RESERVE, PAGE_READWRITE);

// vytvoen prvn strnky pro popisova regionu
			if (region != NULL)
			{
				if (::VirtualAlloc(region, PAGESIZE,
						MEM_COMMIT, PAGE_READWRITE) == NULL)
				{

// pi chyb uvolnn regionu a hlen chyby
					::VirtualFree(region, 0, MEM_RELEASE);
					return FALSE;
				}

// pznak pouit regionu
				*used = TRUE;		// pznak pouit regionu
				ResRegion = region;	// bude te jako rezervn region
				region->Used = 0;		// ta pouitch strnek

// napojen strnek do etzce nepipojench strnek
				page = NULL;
				nextpage = region->Pages;
				for (i = DATAPAGES; i > 0; i--)
				{
					prevpage = page;
					page = nextpage;
					nextpage++;
					page->PrevPage = prevpage;
					page->NextPage = nextpage;
				}
				page->NextPage = FreePage;
				FreePage = region->Pages;
				return TRUE;
			}
		}
	}
	return FALSE;
}


/////////////////////////////////////////////////////////////////////////////
// zruen rezevnho regionu

static void DelRegion()
{
// kontrola, zda je rezervn region
	REGION* region = ResRegion;
	if (region == NULL) return;
	ResRegion = NULL;

// odpojen strnek z etzce nepipojench strnek
	PAGE* page = region->Pages;
	PAGE* nextpage;
	PAGE* prevpage;
	for (int i = DATAPAGES; i > 0; i--)
	{
		nextpage = page->NextPage;
		prevpage = page->PrevPage;

		if (prevpage != NULL)
		{
			prevpage->NextPage = nextpage;
		}
		else
		{
			FreePage = nextpage;
		}

		if (nextpage != NULL)
		{
			nextpage->PrevPage = prevpage;
		}
		page++;
	}

// zruen pznaku pouit regionu
	UsedRegion[(DWORD)region/REGSIZE] = FALSE;

// odpojen strnky popisovae regionu
	VERIFY (::VirtualFree(region, PAGESIZE, MEM_DECOMMIT));

// uvolnn regionu
	VERIFY (::VirtualFree(region, 0, MEM_RELEASE));
}
