
/***************************************************************************\
*																			*
*								Textov etzce								*
*																			*
\***************************************************************************/


/////////////////////////////////////////////////////////////////////////////
// struktura poloky dat etzce (je sdlen mezi vce etzci)

typedef struct STRINGDATA_
{
	long	Refer;					// ta referenc na textov etzec
	int		Length;					// dlka etzce ve znacch
	TCHAR	Data[1];				// textov data
} STRINGDATA;

extern STRINGDATA EmptyStringData;	// data przdnho etzce


/////////////////////////////////////////////////////////////////////////////
// textov etzec

class CString
{
private:
	STRINGDATA*		pData;			// ukazatel na data etzce

public:

// implicitn konstruktor
	inline CString() { Attach(&EmptyStringData); };

// inicializace
	inline void Init() { Attach(&EmptyStringData); };

// deinicializace
	inline void Deinit() { Detach(); };

// destruktor
	~CString();

// koprovac konstruktor
	inline CString(const CString& src) { Attach(src.pData); }

// konverzn konstruktory (-1 = implicitn poet znak)
	CString(LPCTSTR text, int length = -1);
#ifdef _UNICODE
	CString(LPCSTR text, int length = -1);
#else //_UNICODE
	CString(LPCWSTR text, int length = -1);
#endif //_UNICODE

	CString(TCHAR chr);

// opertory
	friend CString operator+ (const CString& string1, const CString& string2);
	friend CString operator+ (LPCTSTR string1, const CString& string2);
	friend CString operator+ (const CString& string1, LPCTSTR string2);
	friend CString operator+ (const TCHAR chr, const CString& string2);
	friend CString operator+ (const CString& string1, const TCHAR chr);
	friend BOOL operator==(const CString& s1, const CString& s2);
	friend BOOL operator==(LPCTSTR s1, const CString& s2);
	friend BOOL operator==(const CString& s1, LPCTSTR s2);
	friend BOOL operator==(const CString& s1, const TCHAR chr);
	friend BOOL operator!=(const CString& s1, const CString& s2);
	friend BOOL operator!=(LPCTSTR s1, const CString& s2);
	friend BOOL operator!=(const CString& s1, LPCTSTR s2);
	friend BOOL operator!=(const CString& s1, const TCHAR chr);

// hledn (-1 = nenalezeno)
	int Find(const CString& src);
	int Find(LPCTSTR text);
	int Find(const TCHAR chr);
	int RevFind(const TCHAR chr);

// konverze mal a velk psmena
	void UpperCase();
	void LowerCase();

// konverze sla INT na text
	void Int(int num);

// konverze textu na slo INT
	friend int Int(LPCTSTR text);

// konverze sla DOUBLE na text
	void Double(double num);

// konverze textu na slo DOUBLE
	friend double Double(LPCTSTR text);

// konverze sla DWORD na HEX (8 slic)
	void Hex(DWORD num);

// pravy textu
	CString Left(int count) const;
	CString Right(int count) const;
	CString Mid(int first, int count = 1) const;
	void Delete(int first, int count = 0x7fffffff);
	void TrimLeft();
	void TrimRight();

// naten textu z resource
	void Load(int nID);

// konverze na etzec LPCTSTR
	inline operator LPCTSTR() const { return pData->Data; }

// poskytnut dlky etzce
	inline int Length() { return pData->Length; }

// vyprzdnn etzce (s ppadnm zruenm bufferu)
	void Empty();

// test etzce, zda je przdn
	inline BOOL IsEmpty() { return pData->Length == 0; };
	inline BOOL IsNotEmpty() { return pData->Length != 0; };

// poskytnut znaku na pozici bez kontroly
	inline TCHAR operator[] (int index) const
	{ 
		ASSERT((DWORD)index < (DWORD)pData->Length);
		return pData->Data[index];
	}

// nastaven znaku na pozici bez kontroly
//	TCHAR& operator[] (int index);

// poskytnut znaku na pozici s kontrolou
	TCHAR Get(int index) const;

// nastaven znaku na pozici s kontrolou
	void Set(int index, TCHAR znak);

// opertory piazen
	const CString& operator= (const CString& src);
	const CString& operator= (LPCTSTR text);
#ifdef _UNICODE
	const CString& operator= (LPCSTR text);
#else //_UNICODE
	const CString& operator= (LPCWSTR text);
#endif //_UNICODE
	const CString& operator= (const TCHAR chr);

// opertory piten
	const CString& operator+=(const CString& src);
	const CString& operator+=(LPCTSTR text);
	const CString& operator+=(const TCHAR chr);

// naten textu okna
	void GetWindowText(const HWND wnd);

// nastaven textu okna
	inline void SetWindowText(const HWND wnd)
	{
		::SetWindowText(wnd, pData->Data);
	}


/////////////////////////////////////////////////////////////////////////////
// privtn funkce

private:

// pipojen dat (pozor - Refer se mus obslouit 1 instrukc !!!)
	inline void Attach(STRINGDATA* data)
	{
		pData = data;
#if defined(_M_IX86) || !defined(_MT)
		data->Refer++;
#else
		InterlockedIncrement(&(data->Refer));
#endif
	}

// odpojen (a zruen) dat (pozor - Refer se mus obslouit 1 instrukc !!!)
	inline void Detach()
	{
		long* refer = &(pData->Refer);

#if defined(_M_IX86) || !defined(_MT)
		if (--(*refer) == 0)			// nemnit na "<= 0" !!!!
#else
		if (InterlockedDecrement(refer) == 0)
#endif
		{
			MemFree(pData);
		}
	}

// kopie do vlastnho bufferu ped modifikac
	inline void CopyWrite()
	{
		STRINGDATA* data = pData;		// adresa starch dat
		long* refer = &(data->Refer);	// poet referenc

		if (*refer > 1)					// je njak jin majitel?
		{
			NewBuffer(data->Length);	// vytvoen novho bufferu
			MemCopy(pData->Data, data->Data, data->Length*sizeof(TCHAR));

// odpojen starch dat - v multithread me nastat i zruen
#if defined(_M_IX86) || !defined(_MT)
			if (--(*refer) == 0)			// nemnit na "<= 0" !!!!
#else
			if (InterlockedDecrement(refer) == 0)
#endif
			{
				MemFree(data);			// ppadn zruen etzce
			}
		}
	}

// vytvoen novho bufferu (dlka zadna ve znacch) - star buffer mus bt odpojen!
	inline void NewBuffer(int length)
	{
		if (length < 0) length = 0;
		int buffsize = length*sizeof(TCHAR) + sizeof(STRINGDATA);
		STRINGDATA* data = (STRINGDATA*)MemGet(buffsize);
		data->Refer = 1;				// poet referenc
		data->Length = length;			// dlka
		data->Data[length] = 0;			// oznaen konce textu
		pData = data;					// adresa dat
	}

// zmna velikosti bufferu - buffer mus bt pivlastnn!
	inline void SizeBuffer(int length)
	{
		ASSERT(pData->Refer == 1);
		if (length < 0) length = 0;
		int buffsize = length*sizeof(TCHAR) + sizeof(STRINGDATA);
		STRINGDATA* data = (STRINGDATA*)MemSize(pData, buffsize);
		data->Length = length;			// dlka
		data->Data[length] = 0;			// oznaen konce textu
		pData = data;					// adresa dat
	}

// sluovac konstruktory (ureno pro opertor +)
	inline CString(const CString& src1, const CString& src2) 
	{ 
		int len1 = src1.pData->Length;
		int len2 = src2.pData->Length;
		NewBuffer(len1+len2);
		MemCopy(pData->Data, src1.pData->Data, len1*sizeof(TCHAR));
		MemCopy(pData->Data + len1, src2.pData->Data, len2*sizeof(TCHAR));
	}

	inline CString(LPCTSTR src1, const CString& src2) 
	{ 
		int len1 = (src1 == NULL) ? 0 : lstrlen(src1);
		int len2 = src2.pData->Length;
		NewBuffer(len1+len2);
		MemCopy(pData->Data, src1, len1*sizeof(TCHAR));
		MemCopy(pData->Data + len1, src2.pData->Data, len2*sizeof(TCHAR));
	}

	inline CString(const CString& src1, LPCTSTR src2) 
	{ 
		int len1 = src1.pData->Length;
		int len2 = (src2 == NULL) ? 0 : lstrlen(src2);
		NewBuffer(len1+len2);
		MemCopy(pData->Data, src1.pData->Data, len1*sizeof(TCHAR));
		MemCopy(pData->Data + len1, src2, len2*sizeof(TCHAR));
	}

	inline CString(const TCHAR src1, const CString& src2) 
	{ 
		int len2 = src2.pData->Length;
		NewBuffer(1+len2);
		pData->Data[0] = src1;
		MemCopy(pData->Data + 1, src2.pData->Data, len2*sizeof(TCHAR));
	}

	inline CString(const CString& src1, const TCHAR src2) 
	{ 
		int len1 = src1.pData->Length;
		NewBuffer(len1+1);
		MemCopy(pData->Data, src1.pData->Data, len1*sizeof(TCHAR));
		pData->Data[len1] = src2;
	}

// porovnn (ureno pro opertor ==)
	inline BOOL Equ(const CString& src) const
	{ 
		int delka = pData->Length;
		if (delka != src.pData->Length) return FALSE;
		return MemCompare(pData->Data, src.pData->Data, delka*sizeof(TCHAR));
	};

	inline BOOL Equ(LPCTSTR src) const
	{ 
		int delka = (src == NULL) ? 0 : lstrlen(src);
		if (delka != pData->Length) return FALSE;
		return MemCompare(pData->Data, src, delka*sizeof(TCHAR));
	};

	inline BOOL Equ(const TCHAR chr) const
	{ 
		if (pData->Length != 1) return FALSE;
		return (pData->Data[0] == chr);
	};

	inline BOOL NEqu(const CString& src) const
	{ 
		int delka = pData->Length;
		if (delka != src.pData->Length) return TRUE;
		return !MemCompare(pData->Data, src.pData->Data, delka*sizeof(TCHAR));
	};

	inline BOOL NEqu(LPCTSTR src) const
	{ 
		int delka = (src == NULL) ? 0 : lstrlen(src);
		if (delka != pData->Length) return TRUE;
		return !MemCompare(pData->Data, src, delka*sizeof(TCHAR));
	};

	inline BOOL NEqu(const TCHAR chr) const
	{ 
		if (pData->Length != 1) return TRUE;
		return (pData->Data[0] != chr);
	};
};
